#include <SPI.h>
#include <time.h>
#include <FS.h>
#include <SD.h>
#include <TFT_eSPI.h>
#include <JPEGDecoder.h>
#include <ArduinoJson.h>
#include <WiFi.h>
#include <HTTPClient.h>
#include <TimeLib.h>

//Enable SPI pins on ESP32
TFT_eSPI tft = TFT_eSPI();

//Assign a pin to the button input
const int buttonPin = 21;

// variables will change:
int buttonState = 0;  

//Readable names for 16-bit color values
#define BLACK   0x0000
#define BLUE    0x001F
#define RED     0xD061
#define GREEN   0x07E0
#define CYAN    0x07FF
#define MAGENTA 0xF81F
#define YELLOW  0xFFE0
#define WHITE   0xFFFF
#define GREY    0xEF7D

//Enter Your WiFi Credentials
const char* ssid = "My_Wi-Fi_Network_Name"; //Enter your SSID
const char* password = "My_Wi-Fi_Password"; //Enter your WiFi password

//Server to request Epoch Time
const char* ntpServer = "pool.ntp.org";
unsigned long epochTime;

int gmtOffset_sec = -14400; //Enter your GMT offset in seconds
int gmtOffset_hr = (gmtOffset_sec/3600); //Converting GMT offset in seconds to hours
const int   daylightOffset_sec = 0; //Enter your daylight saving offset in sec

const String endpoint = "http://api.openweathermap.org/data/2.5/weather?q="; //URL to request weather information
const String city = "New York"; //Enter your city name
const String country = "us"; //Enter your country code
const String lang = "en"; //Enter your language
const String key = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"; //Enter your app-key from OpenWeatherMap.
const String unit = "imperial"; //change to metric or imperial
const String unit_type = "F"; //change to C for metric

//Variables for news API
String key2 = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"; //Enter your app-key from NewsAPI.
const String endpoint2 = "https://newsapi.org/v2/top-headlines?country=" + country + "&apiKey=" + key2 + "&Pagesize=5";

volatile int phase = 0;

void setup() {
  //Initialize Serial
  Serial.begin(115200);
  
  //Declare the button as an input
  pinMode(buttonPin, INPUT);

  //Configure time
  configTime(0, 0, ntpServer);
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);
  printLocalTime();

  //Connect to WiFi
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi..");
  }
  Serial.println("Connected to the WiFi network");

  //Pull Chip Select High
  digitalWrite(15, HIGH); // TFT screen chip select
  digitalWrite( 5, HIGH); // SD card chips select, must use GPIO 5 (ESP32 SS)

  //Initialize SD card
  if (!SD.begin()) {
    Serial.println("Card Failed to Boot");
    return;
  }
  uint8_t cardType = SD.cardType(); 

  if (cardType == CARD_NONE) {
    Serial.println("No SD Card Attached");
    return;
  }
  
  //Initialize the LCD
  tft.begin();
  tft.setRotation(1);
  drawSd("/SMPTE.jpg", 0, 0);
  delay(700);

  //Define the action of the button as an interrupt
  attachInterrupt(digitalPinToInterrupt(21), SMPTE, HIGH);

  //Clearing the display
  tft.fillScreen(BLACK);

}

void loop() {
  
  while (phase == 1) {
    drawSd("/SMPTE.jpg", 0, 0);
    //Check WiFi Availability
    if (WiFi.status() == WL_CONNECTED) {
      HTTPClient http;
      http.begin(endpoint2);
      int httpCode = http.GET();
  
      if (httpCode > 0) {
        String input = http.getString();
        Serial.println(httpCode);
        Serial.println(input);
        DynamicJsonDocument doc(4096);
        deserializeJson(doc, input);
  
        const char* status = doc["status"]; // "ok"
        int totalResults = doc["totalResults"]; // 38
        delay(3000);
        tft.fillScreen(BLACK);
        for (JsonObject articles_item : doc["articles"].as<JsonArray>()) {
  
          const char* articles_item_source_id = articles_item["source"]["id"]; // nullptr, nullptr, ...
          const char* articles_item_source_name = articles_item["source"]["name"]; // "New York Times", "The ...
          const char* articles_item_author = articles_item["author"]; // "John Koblin", "Guardian staff reporter", ...
          String articles_item_title = articles_item["title"]; // "Emmy Awards 2021: What to Watch For - The ...
          String articles_item_description = articles_item["description"]; // "“The Crown” and Netflix could ...
          const char* articles_item_url = articles_item["url"];
          const char* articles_item_urlToImage = articles_item["urlToImage"];
          const char* articles_item_publishedAt = articles_item["publishedAt"]; // "2021-09-19T14:22:31Z", ...
          const char* articles_item_content = articles_item["content"]; // "The Crown already picked up four Emmys ...


          if (articles_item_title == "null") {
            phase = 0;
            break;
          }

        
          Serial.println(articles_item_title);
          Serial.println(articles_item_description);
          Serial.println(articles_item_content);
          Serial.println("---");

          utilities();

          tft.setTextSize(2);
          tft.fillRect(20, 35, 280, 35, RED);
          tft.setTextColor(WHITE, RED);
          tft.drawString("Top Headlines", 82, 35, 2);

          tft.fillRect(20, 65, 280, 175, WHITE);

          int bar = 63; 
          for (int n = 0; n <= 12; n++){
            tft.fillRect(20, bar, 280, 15, GREY);
            bar += 27;       
          }
  
          int gain2 = 30;
          int gain = 0;
          int y_cor = 90;
          tft.setTextSize(1);
          for (int i = 0; i <= 15; i++){
            tft.setTextColor(BLACK, GREY);
            tft.drawString(articles_item_title.substring(gain, gain2), 50, y_cor, 2);
            gain += 30;
            gain2 += 30;
            y_cor += 27;
          }
          delay(8500);
        }
      } 
    }
    tft.fillScreen(BLACK);
    phase = 0;
  }
  if (phase == 0) {
    //Check WiFi Availability
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    http.begin(endpoint + city + "," + country + "&lang=" + lang + "&units=" + unit + "&" + "appid=" + key);
    int httpCode = http.GET();

    //Check for the returning code
    if (httpCode > 0) {
          String input = http.getString();
          Serial.println(httpCode);
          Serial.println(input);
          DynamicJsonDocument doc(1024);
          deserializeJson(doc, input);

          //Initialize all JSON objects
          JsonObject main = doc["main"];
          JsonObject weather_0 = doc["weather"][0];
          JsonObject sys = doc["sys"];

          //Assign a variable to objects in returned JSON 
          const char* weather_0_main = weather_0["main"];
          int main_temp_min = main["temp_min"];
          int main_temp_max = main["temp_max"];
          const char* weather_0_description = weather_0["description"];
          const char* base = doc["base"];
          int main_temp = main["temp"];
          float main_feels_like = main["feels_like"];
          String main_pressure = main["pressure"];
          String main_humidity = main["humidity"];
          int wind_speed = doc["wind"]["speed"];
          int wind_deg = doc["wind"]["deg"];
          int clouds_all = doc["clouds"]["all"];
          long r = sys["sunrise"];
          long s = sys["sunset"];
          const char* sys_country = sys["country"];
          int weather_0_id = weather_0["id"];
          const char* place = doc["name"];

          //Display time
          utilities();

          //Get the date buffer
          char buff[32];
          sprintf(buff, "%02d.%02d.%02d %02d:%02d:%02d", day(r), month(r), year(r), hour(r), minute(r), second(r));
          sprintf(buff, "%02d.%02d.%02d %02d:%02d:%02d", day(s), month(s), year(s), hour(s), minute(s), second(s));
          Serial.println(buff);

          Serial.println(gmtOffset_hr);
          int s_hr = hour(s) + gmtOffset_hr;
          int r_hr = hour(r) + gmtOffset_hr;
          Serial.println(s_hr);
          Serial.println(r_hr);

          //Find the current Epoch time to determine P.M. or A.M
          epochTime = getTime();
          Serial.print("Epoch Time: ");
          Serial.println(epochTime);
          delay(1000);

          //Creating string values for the variables
          String temp = String(main_temp) + "`" + unit_type;
          String temp_max = String(main_temp_max) + "";
          String temp_min = String(main_temp_min) + "";
          String max_min = temp_max + "/" + temp_min + "`" + unit_type;

          //Defining the start point of the variables
          tft.setTextSize(1);
          int startPoint = 320 - (tft.textWidth(place, 4));
          tft.setTextSize(1);
          int startPoint2 = 320 - (tft.textWidth(weather_0_main, 4));
          tft.setTextSize(1);
          int startPoint3 = 320 -(tft.textWidth(max_min, 4));
          tft.setTextSize(2);
          int startPoint4 = 320 - (tft.textWidth(temp, 4));

          //Determine if it is night or day time
          if (epochTime > s or epochTime < r) {
            String NightIconName = "/Night/" + String(weather_0_main) + ".jpg";
            drawSd(NightIconName.c_str(), startPoint - 130, 38);
          }
          if (epochTime < s and epochTime > r) {
            String DayIconName = "/Day/" + String(weather_0_main) + ".jpg";
            drawSd(DayIconName.c_str(), startPoint - 130, 38);
          }

          //Display town name
          tft.setTextSize(1);
          tft.setTextColor(CYAN, BLACK);
          Serial.println(strlen(place));
          tft.drawString(place, startPoint, 33, 4);

          //Display weather description
          tft.setTextSize(1);
          tft.setTextColor(YELLOW, BLACK);
          Serial.println(strlen(weather_0_description));
          tft.drawString(weather_0_main, startPoint2, 58, 4);

          //Display the main temperature
          tft.setTextSize(2);
          tft.setTextColor(WHITE, BLACK);
          tft.drawString(temp, startPoint4, 81, 4); 

          //Display the max & min temperature
          tft.setTextSize(1);
          tft.setTextColor(WHITE, BLACK);
          tft.drawString(max_min, startPoint3, 122, 4);

          //Display humidity percent
          tft.setTextSize(2);
          tft.setTextColor(CYAN, BLACK);
          tft.drawString("H: ", 10, 175, 1);
          tft.setTextColor(WHITE, BLACK);
          tft.drawString(main_humidity + "%", 36, 175, 1);

          //Display pressure
          tft.setTextSize(2);
          tft.setTextColor(CYAN, BLACK);
          tft.drawString("P: ", 10, 200, 1);
          tft.setTextColor(WHITE, BLACK);
          tft.drawString(main_pressure, 36, 200, 1);
          tft.setTextSize(1);
          tft.setTextColor(CYAN, BLACK);
          tft.drawString("hPa", 85, 208);

          //Displaty wind speed
          tft.setTextSize(2);
          tft.setTextColor(CYAN, BLACK);
          tft.drawString("Ws: ", 238, 175, 1);
          tft.setTextColor(WHITE, BLACK);
          tft.drawString(String(wind_speed), 273, 175, 1);
          tft.setTextSize(1);
          tft.setTextColor(CYAN, BLACK);
          tft.drawString("mph", 300, 180);

          //Display wind degree                    
          tft.setTextSize(2);
          tft.setTextColor(CYAN, BLACK);
          tft.drawString("Wd: ", 238, 200, 1);
          tft.setTextColor(WHITE, BLACK);
          tft.drawString(String(wind_deg), 273, 200, 1);

          //Display sunrise time
          tft.setTextSize(2);
          tft.setTextColor(CYAN, BLACK);
          tft.drawString("Rise: ", 117, 175, 1);
          tft.setTextColor(WHITE, BLACK);
          tft.setTextFont(1);
          tft.setCursor(175, 175);
          String r_time = String(r_hr) + ":" + String(minute(r));
          tft.print(r_time);
          Serial.println(r_time);

          //Display sunset time
          tft.setTextSize(2);
          tft.setTextColor(CYAN, BLACK);
          tft.drawString("Set: ", 117, 200, 1);
          tft.setTextColor(WHITE, BLACK);
          tft.setTextFont(1);
          tft.setCursor(164, 200);
          String s_time = String(s_hr) + ":" + String(minute(s));
          tft.print(s_time);
          Serial.println(s_time);

    }
    else {
        Serial.println("Error on HTTP request");
    }
    //End http request
    http.end();
  }
  delay(500);
 }
}

void drawSd(const char *filename, int xpos, int ypos) {

  // Open the named file (the Jpeg decoder library will close it)
  File jpegFile = SD.open(filename, FILE_READ);  // or, file handle reference for SD library
 
  if ( !jpegFile ) {
    Serial.print("ERROR: File \""); Serial.print(filename); Serial.println ("\" not found!");
    return;
  }

  Serial.println("===========================");
  Serial.print("Drawing file: "); Serial.println(filename);
  Serial.println("===========================");

  // Use one of the following methods to initialise the decoder:
  bool decoded = JpegDec.decodeSdFile(jpegFile);  // Pass the SD file handle to the decoder,
  //bool decoded = JpegDec.decodeSdFile(filename);  // or pass the filename (String or character array)

  if (decoded) {
    // print information about the image to the serial port
    jpegInfo();
    // render the image onto the screen at given coordinates
    jpegRender(xpos, ypos);
  }
  else {
    Serial.println("Jpeg file format not supported!");
  }
}

void jpegRender(int xpos, int ypos) {

  //jpegInfo(); // Print information from the JPEG file (could comment this line out)

  uint16_t *pImg;
  uint16_t mcu_w = JpegDec.MCUWidth;
  uint16_t mcu_h = JpegDec.MCUHeight;
  uint32_t max_x = JpegDec.width;
  uint32_t max_y = JpegDec.height;

  bool swapBytes = tft.getSwapBytes();
  tft.setSwapBytes(true);
  
  // Jpeg images are draw as a set of image block (tiles) called Minimum Coding Units (MCUs)
  // Typically these MCUs are 16x16 pixel blocks
  // Determine the width and height of the right and bottom edge image blocks
  uint32_t min_w = jpg_min(mcu_w, max_x % mcu_w);
  uint32_t min_h = jpg_min(mcu_h, max_y % mcu_h);

  // save the current image block size
  uint32_t win_w = mcu_w;
  uint32_t win_h = mcu_h;

  // record the current time so we can measure how long it takes to draw an image
  uint32_t drawTime = millis();

  // save the coordinate of the right and bottom edges to assist image cropping
  // to the screen size
  max_x += xpos;
  max_y += ypos;

  // Fetch data from the file, decode and display
  while (JpegDec.read()) {    // While there is more data in the file
    pImg = JpegDec.pImage ;   // Decode a MCU (Minimum Coding Unit, typically a 8x8 or 16x16 pixel block)

    // Calculate coordinates of top left corner of current MCU
    int mcu_x = JpegDec.MCUx * mcu_w + xpos;
    int mcu_y = JpegDec.MCUy * mcu_h + ypos;

    // check if the image block size needs to be changed for the right edge
    if (mcu_x + mcu_w <= max_x) win_w = mcu_w;
    else win_w = min_w;

    // check if the image block size needs to be changed for the bottom edge
    if (mcu_y + mcu_h <= max_y) win_h = mcu_h;
    else win_h = min_h;

    // copy pixels into a contiguous block
    if (win_w != mcu_w)
    {
      uint16_t *cImg;
      int p = 0;
      cImg = pImg + win_w;
      for (int h = 1; h < win_h; h++)
      {
        p += mcu_w;
        for (int w = 0; w < win_w; w++)
        {
          *cImg = *(pImg + w + p);
          cImg++;
        }
      }
    }

    // calculate how many pixels must be drawn
    uint32_t mcu_pixels = win_w * win_h;

    // draw image MCU block only if it will fit on the screen
    if (( mcu_x + win_w ) <= tft.width() && ( mcu_y + win_h ) <= tft.height())
      tft.pushImage(mcu_x, mcu_y, win_w, win_h, pImg);
    else if ( (mcu_y + win_h) >= tft.height())
      JpegDec.abort(); // Image has run off bottom of screen so abort decoding
  }

  tft.setSwapBytes(swapBytes);

  showTime(millis() - drawTime); // These lines are for sketch testing only
}

void jpegInfo() {

  // Print information extracted from the JPEG file
  Serial.println("JPEG image info");
  Serial.println("===============");
  Serial.print("Width      :");
  Serial.println(JpegDec.width);
  Serial.print("Height     :");
  Serial.println(JpegDec.height);
  Serial.print("Components :");
  Serial.println(JpegDec.comps);
  Serial.print("MCU / row  :");
  Serial.println(JpegDec.MCUSPerRow);
  Serial.print("MCU / col  :");
  Serial.println(JpegDec.MCUSPerCol);
  Serial.print("Scan type  :");
  Serial.println(JpegDec.scanType);
  Serial.print("MCU width  :");
  Serial.println(JpegDec.MCUWidth);
  Serial.print("MCU height :");
  Serial.println(JpegDec.MCUHeight);
  Serial.println("===============");
  Serial.println("");
}

void showTime(uint32_t msTime) {
  //tft.setCursor(0, 0);
  //tft.setTextFont(1);
  //tft.setTextSize(2);
  //tft.setTextColor(TFT_WHITE, TFT_BLACK);
  //tft.print(F(" JPEG drawn in "));
  //tft.print(msTime);
  //tft.println(F(" ms "));
  Serial.print(F(" JPEG drawn in "));
  Serial.print(msTime);
  Serial.println(F(" ms "));

}

unsigned long getTime() {
  time_t now;
  struct tm timeinfo;
  if (!getLocalTime(&timeinfo)) {
    //Serial.println("Failed to obtain time");
    return(0);
  }
  time(&now);
  return now;
}

void printLocalTime(){
  struct tm timeinfo;
  if(!getLocalTime(&timeinfo)){
    Serial.println("Failed to obtain time");
    return;
  }
  Serial.println(&timeinfo, "%A, %B %d %Y %H:%M:%S");
  Serial.print("Day of week: ");
  Serial.println(&timeinfo, "%A");
  Serial.print("Month: ");
  Serial.println(&timeinfo, "%B");
  Serial.print("Day of Month: ");
  Serial.println(&timeinfo, "%d");
  Serial.print("Year: ");
  Serial.println(&timeinfo, "%Y");
  Serial.print("Hour: ");
  Serial.println(&timeinfo, "%H");
  Serial.print("Hour (12 hour format): ");
  Serial.println(&timeinfo, "%I");
  Serial.print("Minute: ");
  Serial.println(&timeinfo, "%M");
  Serial.print("Second: ");
  Serial.println(&timeinfo, "%S");

  Serial.println("Time variables");
  char timeHour[3];
  strftime(timeHour,3, "%H", &timeinfo);
  Serial.println(timeHour);
  char timeWeekDay[10];
  strftime(timeWeekDay,10, "%A", &timeinfo);
  Serial.println(timeWeekDay);
  Serial.println();
}

void utilities(){
  
  struct tm timeinfo;
  if(!getLocalTime(&timeinfo)){
    Serial.println("Failed to obtain time");
    return;
  }
  
  tft.setTextColor(WHITE, BLACK);  
  tft.setCursor(140, 7);
  tft.setTextSize(2);
  tft.print(&timeinfo, "%H");
  tft.setCursor(162, 7);
  tft.setTextSize(2);
  tft.print(":");
  tft.setCursor(171, 7);
  tft.setTextSize(2);
  tft.print(&timeinfo, "%M");

}
void SMPTE() {
  phase++;
}
